package ru.dbotthepony.kstarbound.lua

import it.unimi.dsi.fastutil.ints.IntAVLTreeSet
import java.util.concurrent.ConcurrentLinkedQueue

class LuaHandleThread(mainThread: LuaThread) {
	private val pendingFree = ConcurrentLinkedQueue<Int>()
	private val freeHandles = IntAVLTreeSet()
	private var nextHandle = 0
	// faster code path
	private var handlesInUse = 0

	val thread = mainThread.newThread(true)

	init {
		mainThread.storeRef(LuaThread.LUA_REGISTRYINDEX)
	}

	fun freeHandle(handle: Int) {
		pendingFree.add(handle)
	}

	fun cleanup() {
		if (handlesInUse == 0) return
		var handle = pendingFree.poll()

		while (handle != null) {
			handlesInUse--
			freeHandles.add(handle)
			thread.push()
			thread.copy(-1, handle)
			thread.pop()

			handle = pendingFree.poll()
		}
	}

	fun allocateHandle(): LuaHandle {
		handlesInUse++

		if (freeHandles.isEmpty()) {
			return LuaHandle(this, ++nextHandle)
		} else {
			val handle = freeHandles.firstInt()
			freeHandles.remove(handle)

			thread.copy(-1, handle)
			thread.pop()
			return LuaHandle(this, handle)
		}
	}
}